from setuptools import setup, find_packages

# here = pathlib.Path(__file__).parent.resolve()
print('kewl!!!!!!!!!')
if __name__ == "__main__":
    setup()
# setup(
#     name='plankton',
#     version='0.1.0',
#     description='A point process analysis project.',
#     package_dir={'': 'plankton'},
#     packages=find_packages('plankton'),
#     python_requires=">3.7",
#     install_requires=['numpy',
#                       'anndata',
#                       'scanpy',
#                       'jupyter',
#                       'matplotlib',
#                       'numpy',
#                       'pandas',
#                       'plotly',
#                       'scipy',
#                       'scikit-image',
#                       'scikit-learn',
#                       'umap-learn',
#                       ]
# )
